@@ -8,8 +8,6 @@
 
 #include <coreplugin/progressmanager/progressmanager.h>
 
-#include <extensionsystem/pluginmanager.h>
-
 #include <projectexplorer/treescanner.h>
 
 #include <utils/async.h>
@@ -132,6 +130,15 @@ CompilationDbParser::CompilationDbParser(const QString &projectName,
     });
 }
 
+CompilationDbParser::~CompilationDbParser()
+{
+    if (m_treeScanner && !m_treeScanner->isFinished()) {
+        auto future = m_treeScanner->future();
+        future.cancel();
+        future.waitForFinished();
+    }
+}
+
 void CompilationDbParser::start()
 {
     // Check hash first.
@@ -160,12 +167,17 @@ void CompilationDbParser::start()
 
             // Cache mime check result for speed up
             if (!isIgnored) {
-                auto it = m_mimeBinaryCache.find(mimeType.name());
-                if (it != m_mimeBinaryCache.end()) {
+                if (auto it = m_mimeBinaryCache.get<std::optional<bool>>(
+                        [mimeType](const QHash<QString, bool> &cache) -> std::optional<bool> {
+                            const auto cache_it = cache.find(mimeType.name());
+                            if (cache_it != cache.end())
+                                return *cache_it;
+                            return {};
+                        })) {
                     isIgnored = *it;
                 } else {
                     isIgnored = TreeScanner::isMimeBinary(mimeType, fn);
-                    m_mimeBinaryCache[mimeType.name()] = isIgnored;
+                    m_mimeBinaryCache.writeLocked()->insert(mimeType.name(), isIgnored);
                 }
             }
 
@@ -190,7 +202,7 @@ void CompilationDbParser::start()
                                    "CompilationDatabase.Parse");
     ++m_runningParserJobs;
     m_parserWatcher.setFuture(future);
-    ExtensionSystem::PluginManager::futureSynchronizer()->addFuture(future);
+    Utils::futureSynchronizer()->addFuture(future);
 }
 
 void CompilationDbParser::stop()
