@@ -1,30 +1,7 @@
-/****************************************************************************
-**
-** Copyright (C) 2016 The Qt Company Ltd.
-** Contact: https://www.qt.io/licensing/
-**
-** This file is part of Qt Creator.
-**
-** Commercial License Usage
-** Licensees holding valid commercial Qt licenses may use this file in
-** accordance with the commercial license agreement provided with the
-** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see https://www.qt.io/terms-conditions. For further
-** information use the contact form at https://www.qt.io/contact-us.
-**
-** GNU General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3 as published by the Free Software
-** Foundation with exceptions as appearing in the file LICENSE.GPL3-EXCEPT
-** included in the packaging of this file. Please review the following
-** information to ensure the GNU General Public License requirements will
-** be met: https://www.gnu.org/licenses/gpl-3.0.html.
-**
-****************************************************************************/
+// Copyright (C) 2016 The Qt Company Ltd.
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only WITH Qt-GPL-exception-1.0
 
 #include "syntaxhighlighter.h"
-#include "textdocument.h"
 #include "textdocumentlayout.h"
 #include "texteditorsettings.h"
 #include "fontsettings.h"
@@ -32,28 +9,36 @@
 #include <utils/algorithm.h>
 #include <utils/qtcassert.h>
 
-#include <QTextDocument>
 #include <QPointer>
-#include <qtimer.h>
+#include <QTextDocument>
+#include <QThread>
 
-#include <math.h>
+#include <cmath>
 
 namespace TextEditor {
 
+enum HighlighterTypeProperty
+{
+    SyntaxHighlight = QTextFormat::UserProperty + 1,
+    SemanticHighlight = QTextFormat::UserProperty + 2
+};
+
 class SyntaxHighlighterPrivate
 {
-    SyntaxHighlighter *q_ptr = nullptr;
-    Q_DECLARE_PUBLIC(SyntaxHighlighter)
 public:
-    SyntaxHighlighterPrivate()
+    SyntaxHighlighterPrivate() = default;
+
+    SyntaxHighlighterPrivate(const FontSettings &fontSettings)
     {
-        updateFormats(TextEditorSettings::fontSettings());
+        updateFormats(fontSettings);
     }
 
     QPointer<QTextDocument> doc;
 
+    void updateFormats(int from, int charsRemoved, int charsAdded);
     void reformatBlocks(int from, int charsRemoved, int charsAdded);
-    void reformatBlock(const QTextBlock &block, int from, int charsRemoved, int charsAdded);
+    void reformatBlocks();
+    void reformatBlock(const QTextBlock &block);
 
     inline void rehighlight(QTextCursor &cursor, QTextCursor::MoveOperation operation) {
         inReformatBlocks = true;
@@ -63,27 +48,33 @@ public:
         inReformatBlocks = false;
     }
 
-    void applyFormatChanges(int from, int charsRemoved, int charsAdded);
+    void applyFormatChanges();
     void updateFormats(const FontSettings &fontSettings);
 
-    QVector<QTextCharFormat> formatChanges;
+    FontSettings fontSettings;
+    QList<QTextCharFormat> formatChanges;
     QTextBlock currentBlock;
     bool rehighlightPending = false;
     bool inReformatBlocks = false;
     TextDocumentLayout::FoldValidator foldValidator;
-    QVector<QTextCharFormat> formats;
-    QVector<std::pair<int,TextStyle>> formatCategories;
+    QList<QTextCharFormat> formats;
+    QList<std::pair<int,TextStyle>> formatCategories;
     QTextCharFormat whitespaceFormat;
-    bool noAutomaticHighlighting = false;
+    QString mimeType;
+    bool syntaxInfoUpToDate = false;
+    int highlightStartBlock = 0;
+    int highlightEndBlock = 0;
+    QSet<int> forceRehighlightBlocks;
+    SyntaxHighlighter *q;
 };
 
-static bool adjustRange(QTextLayout::FormatRange &range, int from, int charsRemoved, int charsAdded) {
-
+static bool adjustRange(QTextLayout::FormatRange &range, int from, int charsDelta)
+{
     if (range.start >= from) {
-        range.start += charsAdded - charsRemoved;
+        range.start += charsDelta;
         return true;
     } else if (range.start + range.length > from) {
-        range.length += charsAdded - charsRemoved;
+        range.length += charsDelta;
         return true;
     }
     return false;
@@ -91,45 +82,44 @@ static bool adjustRange(QTextLayout::FormatRange &range, int from, int charsRemo
 
 void SyntaxHighlighter::delayedRehighlight()
 {
-    Q_D(SyntaxHighlighter);
     if (!d->rehighlightPending)
         return;
     d->rehighlightPending = false;
+
     rehighlight();
 }
 
-void SyntaxHighlighterPrivate::applyFormatChanges(int from, int charsRemoved, int charsAdded)
+void SyntaxHighlighter::continueRehighlight()
 {
-    bool formatsChanged = false;
+    d->reformatBlocks();
+}
 
+#ifdef WITH_TESTS
+SyntaxHighlighter::SyntaxHighlighter(QTextDocument *parent, const FontSettings &fontsettings)
+    : QObject(parent), d(new SyntaxHighlighterPrivate(fontsettings))
+{
+    d->q = this;
+    if (parent)
+        setDocument(parent);
+}
+#endif
+
+void SyntaxHighlighterPrivate::applyFormatChanges()
+{
     QTextLayout *layout = currentBlock.layout();
 
-    QVector<QTextLayout::FormatRange> ranges = layout->formats();
-
-    bool doAdjustRange = currentBlock.contains(from);
-
-    QVector<QTextLayout::FormatRange> old_ranges;
-
-    if (!ranges.isEmpty()) {
-        auto it = ranges.begin();
-        while (it != ranges.end()) {
-            if (it->format.property(QTextFormat::UserProperty).toBool()) {
-                if (doAdjustRange)
-                    formatsChanged = adjustRange(*it, from - currentBlock.position(), charsRemoved, charsAdded)
-                            || formatsChanged;
-                ++it;
-            } else {
-                old_ranges.append(*it);
-                it = ranges.erase(it);
-            }
-        }
-    }
+    QList<QTextLayout::FormatRange> ranges;
+    QList<QTextLayout::FormatRange> oldRanges;
+    std::tie(oldRanges, ranges)
+        = Utils::partition(layout->formats(), [](const QTextLayout::FormatRange &range) {
+              return range.format.property(SyntaxHighlight).toBool();
+          });
 
     QTextCharFormat emptyFormat;
 
     QTextLayout::FormatRange r;
 
-    QVector<QTextLayout::FormatRange> new_ranges;
+    QList<QTextLayout::FormatRange> newRanges;
     int i = 0;
     while (i < formatChanges.count()) {
 
@@ -145,21 +135,36 @@ void SyntaxHighlighterPrivate::applyFormatChanges(int from, int charsRemoved, in
         while (i < formatChanges.count() && formatChanges.at(i) == r.format)
             ++i;
 
+        r.format.setProperty(SyntaxHighlight, true);
         r.length = i - r.start;
 
-        new_ranges << r;
+        const QString preeditText = currentBlock.layout()->preeditAreaText();
+        if (!preeditText.isEmpty()) {
+            const int preeditPosition = currentBlock.layout()->preeditAreaPosition();
+            if (r.start >= preeditPosition) {
+                r.start += preeditText.length();
+            } else if (r.start + r.length > preeditPosition) {
+                QTextLayout::FormatRange beforePreeditRange = r;
+                r.start = preeditPosition + preeditText.length();
+                r.length = r.length - (r.start - preeditPosition);
+                beforePreeditRange.length = preeditPosition - beforePreeditRange.start;
+                newRanges << beforePreeditRange;
+            }
+        }
+
+        newRanges << r;
     }
 
-    formatsChanged = formatsChanged || (new_ranges.size() != old_ranges.size());
+    bool formatsChanged = (newRanges.size() != oldRanges.size());
 
-    for (int i = 0; !formatsChanged && i < new_ranges.size(); ++i) {
-        const QTextLayout::FormatRange &o = old_ranges.at(i);
-        const QTextLayout::FormatRange &n = new_ranges.at(i);
+    for (int i = 0; !formatsChanged && i < newRanges.size(); ++i) {
+        const QTextLayout::FormatRange &o = oldRanges.at(i);
+        const QTextLayout::FormatRange &n = newRanges.at(i);
         formatsChanged = (o.start != n.start || o.length != n.length || o.format != n.format);
     }
 
     if (formatsChanged) {
-        ranges.append(new_ranges);
+        ranges.append(newRanges);
         layout->setFormats(ranges);
         doc->markContentsDirty(currentBlock.position(), currentBlock.length());
     }
@@ -167,56 +172,103 @@ void SyntaxHighlighterPrivate::applyFormatChanges(int from, int charsRemoved, in
 
 void SyntaxHighlighter::reformatBlocks(int from, int charsRemoved, int charsAdded)
 {
-    Q_D(SyntaxHighlighter);
     if (!d->inReformatBlocks)
         d->reformatBlocks(from, charsRemoved, charsAdded);
 }
 
+void SyntaxHighlighterPrivate::updateFormats(int from, int charsRemoved, int charsAdded)
+{
+    bool formatsChanged = false;
+
+    const QTextBlock block = doc->findBlock(from);
+    QTextLayout *layout = block.layout();
+
+    QList<QTextLayout::FormatRange> ranges = layout->formats();
+
+    const int charsDelta = charsAdded - charsRemoved;
+    for (QTextLayout::FormatRange &range : ranges)
+        formatsChanged |= adjustRange(range, from - block.position(), charsDelta);
+
+    if (formatsChanged) {
+        layout->setFormats(ranges);
+        doc->markContentsDirty(block.position(), block.length());
+    }
+}
+
 void SyntaxHighlighterPrivate::reformatBlocks(int from, int charsRemoved, int charsAdded)
 {
-    foldValidator.reset();
-
-    rehighlightPending = false;
+    updateFormats(from, charsRemoved, charsAdded);
 
     QTextBlock block = doc->findBlock(from);
+    if (block.isValid() && block.blockNumber() < highlightStartBlock)
+        highlightStartBlock = block.blockNumber();
+    block = doc->findBlock(from + charsAdded + (charsRemoved > 0 ? 1 : 0));
     if (!block.isValid())
-        return;
+        highlightEndBlock = doc->blockCount() - 1;
+    else if (block.blockNumber() > highlightEndBlock)
+        highlightEndBlock = block.blockNumber();
 
-    int endPosition;
-    QTextBlock lastBlock = doc->findBlock(from + charsAdded + (charsRemoved > 0 ? 1 : 0));
-    if (lastBlock.isValid())
-        endPosition = lastBlock.position() + lastBlock.length();
-    else
-        endPosition =  doc->lastBlock().position() + doc->lastBlock().length(); //doc->docHandle()->length();
+    reformatBlocks();
+}
+
+void SyntaxHighlighterPrivate::reformatBlocks()
+{
+    QElapsedTimer et;
+    et.start();
+
+    syntaxInfoUpToDate = false;
+    rehighlightPending = false;
+
+    foldValidator.reset();
 
     bool forceHighlightOfNextBlock = false;
+    QTextBlock block = doc->findBlockByNumber(highlightStartBlock);
+    QTC_ASSERT(block.isValid(), block = doc->firstBlock());
+    QTextBlock endBlock = doc->findBlockByNumber(highlightEndBlock);
+    QTC_ASSERT(endBlock.isValid(), endBlock = doc->lastBlock());
+
+    while (block.isValid()) {
+        if (et.elapsed() > 20)
+            break;
 
-    while (block.isValid() && (block.position() < endPosition || forceHighlightOfNextBlock)) {
         const int stateBeforeHighlight = block.userState();
 
-        reformatBlock(block, from, charsRemoved, charsAdded);
-
-        forceHighlightOfNextBlock = (block.userState() != stateBeforeHighlight);
+        if (forceHighlightOfNextBlock || forceRehighlightBlocks.contains(block.blockNumber())
+                || block.blockNumber() <= highlightEndBlock) {
+            reformatBlock(block);
+            forceRehighlightBlocks.remove(block.blockNumber());
+            forceHighlightOfNextBlock = (block.userState() != stateBeforeHighlight);
+        }
+        highlightStartBlock = block.blockNumber();
 
+        if (block == endBlock && !forceHighlightOfNextBlock)
+            break;
         block = block.next();
     }
 
     formatChanges.clear();
-
     foldValidator.finalize();
+
+    if (endBlock.isValid() && block.isValid() && block.blockNumber() < endBlock.blockNumber()) {
+        QMetaObject::invokeMethod(q, &SyntaxHighlighter::continueRehighlight, Qt::QueuedConnection);
+        if (forceHighlightOfNextBlock)
+            forceRehighlightBlocks << block.blockNumber();
+    } else {
+        highlightEndBlock = 0;
+        syntaxInfoUpToDate = true;
+        emit q->finished();
+    }
 }
 
-void SyntaxHighlighterPrivate::reformatBlock(const QTextBlock &block, int from, int charsRemoved, int charsAdded)
+void SyntaxHighlighterPrivate::reformatBlock(const QTextBlock &block)
 {
-    Q_Q(SyntaxHighlighter);
-
-    Q_ASSERT_X(!currentBlock.isValid(), "SyntaxHighlighter::reformatBlock()", "reFormatBlock() called recursively");
+    QTC_ASSERT(!currentBlock.isValid(), return);
 
     currentBlock = block;
 
     formatChanges.fill(QTextCharFormat(), block.length() - 1);
     q->highlightBlock(block.text());
-    applyFormatChanges(from, charsRemoved, charsAdded);
+    applyFormatChanges();
 
     foldValidator.process(currentBlock);
 
@@ -257,9 +309,9 @@ void SyntaxHighlighterPrivate::reformatBlock(const QTextBlock &block, int from,
     Constructs a SyntaxHighlighter with the given \a parent.
 */
 SyntaxHighlighter::SyntaxHighlighter(QObject *parent)
-    : QObject(parent), d_ptr(new SyntaxHighlighterPrivate)
+    : QObject(parent), d(new SyntaxHighlighterPrivate)
 {
-    d_ptr->q_ptr = this;
+    d->q = this;
 }
 
 /*!
@@ -268,9 +320,9 @@ SyntaxHighlighter::SyntaxHighlighter(QObject *parent)
     SyntaxHighlighter.
 */
 SyntaxHighlighter::SyntaxHighlighter(QTextDocument *parent)
-    : QObject(parent), d_ptr(new SyntaxHighlighterPrivate)
+    : QObject(parent), d(new SyntaxHighlighterPrivate)
 {
-    d_ptr->q_ptr = this;
+    d->q = this;
     if (parent)
         setDocument(parent);
 }
@@ -281,9 +333,9 @@ SyntaxHighlighter::SyntaxHighlighter(QTextDocument *parent)
     the SyntaxHighlighter.
 */
 SyntaxHighlighter::SyntaxHighlighter(QTextEdit *parent)
-    : QObject(parent), d_ptr(new SyntaxHighlighterPrivate)
+    : QObject(parent), d(new SyntaxHighlighterPrivate)
 {
-    d_ptr->q_ptr = this;
+    d->q = this;
     if (parent)
         setDocument(parent->document());
 }
@@ -293,7 +345,7 @@ SyntaxHighlighter::SyntaxHighlighter(QTextEdit *parent)
 */
 SyntaxHighlighter::~SyntaxHighlighter()
 {
-    setDocument(0);
+    setDocument(nullptr);
 }
 
 /*!
@@ -302,7 +354,9 @@ SyntaxHighlighter::~SyntaxHighlighter()
 */
 void SyntaxHighlighter::setDocument(QTextDocument *doc)
 {
-    Q_D(SyntaxHighlighter);
+    if (d->doc == doc)
+        return;
+
     if (d->doc) {
         disconnect(d->doc, &QTextDocument::contentsChange, this, &SyntaxHighlighter::reformatBlocks);
 
@@ -312,13 +366,12 @@ void SyntaxHighlighter::setDocument(QTextDocument *doc)
             blk.layout()->clearFormats();
         cursor.endEditBlock();
     }
+    QTextDocument *oldDoc = d->doc;
     d->doc = doc;
+    documentChanged(oldDoc, d->doc);
     if (d->doc) {
-        if (!d->noAutomaticHighlighting) {
-            connect(d->doc, &QTextDocument::contentsChange, this, &SyntaxHighlighter::reformatBlocks);
-            d->rehighlightPending = true;
-            QTimer::singleShot(0, this, &SyntaxHighlighter::delayedRehighlight);
-        }
+        connect(d->doc, &QTextDocument::contentsChange, this, &SyntaxHighlighter::reformatBlocks);
+        scheduleRehighlight();
         d->foldValidator.setup(qobject_cast<TextDocumentLayout *>(doc->documentLayout()));
     }
 }
@@ -329,10 +382,19 @@ void SyntaxHighlighter::setDocument(QTextDocument *doc)
 */
 QTextDocument *SyntaxHighlighter::document() const
 {
-    Q_D(const SyntaxHighlighter);
     return d->doc;
 }
 
+void SyntaxHighlighter::setMimeType(const QString &mimeType)
+{
+    d->mimeType = mimeType;
+}
+
+QString SyntaxHighlighter::mimeType() const
+{
+    return d->mimeType;
+}
+
 /*!
     \since 4.2
 
@@ -342,7 +404,6 @@ QTextDocument *SyntaxHighlighter::document() const
 */
 void SyntaxHighlighter::rehighlight()
 {
-    Q_D(SyntaxHighlighter);
     if (!d->doc)
         return;
 
@@ -350,6 +411,17 @@ void SyntaxHighlighter::rehighlight()
     d->rehighlight(cursor, QTextCursor::End);
 }
 
+void SyntaxHighlighter::scheduleRehighlight()
+{
+    if (d->rehighlightPending)
+        return;
+    d->rehighlightPending = true;
+    d->syntaxInfoUpToDate = false;
+    QMetaObject::invokeMethod(this,
+                              &SyntaxHighlighter::delayedRehighlight,
+                              Qt::QueuedConnection);
+}
+
 /*!
     \since 4.6
 
@@ -359,7 +431,6 @@ void SyntaxHighlighter::rehighlight()
 */
 void SyntaxHighlighter::rehighlightBlock(const QTextBlock &block)
 {
-    Q_D(SyntaxHighlighter);
     if (!d->doc || !block.isValid() || block.document() != d->doc)
         return;
 
@@ -433,7 +504,6 @@ void SyntaxHighlighter::rehighlightBlock(const QTextBlock &block)
 */
 void SyntaxHighlighter::setFormat(int start, int count, const QTextCharFormat &format)
 {
-    Q_D(SyntaxHighlighter);
     if (start < 0 || start >= d->formatChanges.count())
         return;
 
@@ -480,9 +550,8 @@ void SyntaxHighlighter::setFormat(int start, int count, const QFont &font)
 
 void SyntaxHighlighter::formatSpaces(const QString &text, int start, int count)
 {
-    Q_D(const SyntaxHighlighter);
     int offset = start;
-    const int end = std::min(start + count, text.length());
+    const int end = std::min(start + count, int(text.length()));
     while (offset < end) {
         if (text.at(offset).isSpace()) {
             int start = offset++;
@@ -506,11 +575,9 @@ void SyntaxHighlighter::formatSpaces(const QString &text, int start, int count)
 void SyntaxHighlighter::setFormatWithSpaces(const QString &text, int start, int count,
                                             const QTextCharFormat &format)
 {
-    Q_D(const SyntaxHighlighter);
-    QTextCharFormat visualSpaceFormat = d->whitespaceFormat;
-    visualSpaceFormat.setBackground(format.background());
+    const QTextCharFormat visualSpaceFormat = whitespacified(format);
 
-    const int end = std::min(start + count, text.length());
+    const int end = std::min(start + count, int(text.length()));
     int index = start;
 
     while (index != end) {
@@ -534,10 +601,7 @@ void SyntaxHighlighter::setFormatWithSpaces(const QString &text, int start, int
 */
 QTextCharFormat SyntaxHighlighter::format(int pos) const
 {
-    Q_D(const SyntaxHighlighter);
-    if (pos < 0 || pos >= d->formatChanges.count())
-        return QTextCharFormat();
-    return d->formatChanges.at(pos);
+    return d->formatChanges.value(pos);
 }
 
 /*!
@@ -549,7 +613,6 @@ QTextCharFormat SyntaxHighlighter::format(int pos) const
 */
 int SyntaxHighlighter::previousBlockState() const
 {
-    Q_D(const SyntaxHighlighter);
     if (!d->currentBlock.isValid())
         return -1;
 
@@ -566,7 +629,6 @@ int SyntaxHighlighter::previousBlockState() const
 */
 int SyntaxHighlighter::currentBlockState() const
 {
-    Q_D(const SyntaxHighlighter);
     if (!d->currentBlock.isValid())
         return -1;
 
@@ -580,7 +642,6 @@ int SyntaxHighlighter::currentBlockState() const
 */
 void SyntaxHighlighter::setCurrentBlockState(int newState)
 {
-    Q_D(SyntaxHighlighter);
     if (!d->currentBlock.isValid())
         return;
 
@@ -623,7 +684,6 @@ void SyntaxHighlighter::setCurrentBlockState(int newState)
 */
 void SyntaxHighlighter::setCurrentBlockUserData(QTextBlockUserData *data)
 {
-    Q_D(SyntaxHighlighter);
     if (!d->currentBlock.isValid())
         return;
 
@@ -638,9 +698,8 @@ void SyntaxHighlighter::setCurrentBlockUserData(QTextBlockUserData *data)
 */
 QTextBlockUserData *SyntaxHighlighter::currentBlockUserData() const
 {
-    Q_D(const SyntaxHighlighter);
     if (!d->currentBlock.isValid())
-        return 0;
+        return nullptr;
 
     return d->currentBlock.userData();
 }
@@ -652,7 +711,6 @@ QTextBlockUserData *SyntaxHighlighter::currentBlockUserData() const
 */
 QTextBlock SyntaxHighlighter::currentBlock() const
 {
-    Q_D(const SyntaxHighlighter);
     return d->currentBlock;
 }
 
@@ -661,71 +719,95 @@ static bool byStartOfRange(const QTextLayout::FormatRange &range, const QTextLay
     return range.start < other.start;
 }
 
-// The formats is passed in by reference in order to prevent unnecessary copying of its items.
-// After this function returns, the list is modified, and should be considered invalidated!
 void SyntaxHighlighter::setExtraFormats(const QTextBlock &block,
-                                        QVector<QTextLayout::FormatRange> &formats)
+                                        const QList<QTextLayout::FormatRange> &formats)
 {
-//    qDebug() << "setFormats() on block" << block.blockNumber();
-//    qDebug() << "   is valid:" << (block.isValid() ? "Yes" : "No");
-//    qDebug() << "   has layout:" << (block.layout() ? "Yes" : "No");
-//    if (block.layout()) qDebug() << "   has text:" << (block.text().isEmpty() ? "No" : "Yes");
-
-//    for (int i = 0; i < overrides.count(); ++i)
-//        qDebug() << "   from " << overrides.at(i).start << "length"
-//                 << overrides.at(i).length
-//                 << "color:" << overrides.at(i).format.foreground().color();
-    Q_D(SyntaxHighlighter);
+    QList<QTextLayout::FormatRange> formatsCopy = formats;
 
     const int blockLength = block.length();
-    if (block.layout() == 0 || blockLength == 0)
+    if (block.layout() == nullptr || blockLength == 0)
         return;
 
-    Utils::sort(formats, byStartOfRange);
-
-    const QVector<QTextLayout::FormatRange> all = block.layout()->formats();
-    QVector<QTextLayout::FormatRange> previousSemanticFormats;
-    QVector<QTextLayout::FormatRange> formatsToApply;
-    previousSemanticFormats.reserve(all.size());
-    formatsToApply.reserve(all.size() + formats.size());
-
-    for (int i = 0, ei = formats.size(); i < ei; ++i)
-        formats[i].format.setProperty(QTextFormat::UserProperty, true);
-
-    foreach (const QTextLayout::FormatRange &r, all) {
-        if (r.format.hasProperty(QTextFormat::UserProperty))
-            previousSemanticFormats.append(r);
-        else
-            formatsToApply.append(r);
+    const QString preeditText = block.layout()->preeditAreaText();
+    if (!preeditText.isEmpty()) {
+        QList<QTextLayout::FormatRange> additionalRanges;
+        const int preeditPosition = block.layout()->preeditAreaPosition();
+        for (QTextLayout::FormatRange &r : formatsCopy) {
+            if (r.start >= preeditPosition) {
+                r.start += preeditText.length();
+            } else if (r.start + r.length > preeditPosition) {
+                QTextLayout::FormatRange afterPreeditRange = r;
+                afterPreeditRange.start = preeditPosition + preeditText.length();
+                afterPreeditRange.length = r.length - (preeditPosition - r.start);
+                additionalRanges << afterPreeditRange;
+                r.length = preeditPosition - r.start;
+            }
+        }
+        formatsCopy << additionalRanges;
     }
 
-    if (formats.size() == previousSemanticFormats.size()) {
+    Utils::sort(formatsCopy, byStartOfRange);
+
+    const QList<QTextLayout::FormatRange> all = block.layout()->formats();
+    QList<QTextLayout::FormatRange> previousSemanticFormats;
+    QList<QTextLayout::FormatRange> formatsToApply;
+    std::tie(previousSemanticFormats, formatsToApply)
+        = Utils::partition(all, [](const QTextLayout::FormatRange &r) {
+              return r.format.property(SemanticHighlight).toBool();
+          });
+
+    for (auto &format : formatsCopy)
+        format.format.setProperty(SemanticHighlight, true);
+
+    if (formatsCopy.size() == previousSemanticFormats.size()) {
         Utils::sort(previousSemanticFormats, byStartOfRange);
-
-        int index = 0;
-        for (; index != formats.size(); ++index) {
-            const QTextLayout::FormatRange &range = formats.at(index);
-            const QTextLayout::FormatRange &previousRange = previousSemanticFormats.at(index);
-
-            if (range.start != previousRange.start ||
-                    range.length != previousRange.length ||
-                    range.format != previousRange.format)
-                break;
-        }
-
-        if (index == formats.size())
+        if (formats == previousSemanticFormats)
             return;
     }
 
-    formatsToApply += formats;
+    formatsToApply += formatsCopy;
 
     bool wasInReformatBlocks = d->inReformatBlocks;
     d->inReformatBlocks = true;
     block.layout()->setFormats(formatsToApply);
+
     document()->markContentsDirty(block.position(), blockLength - 1);
     d->inReformatBlocks = wasInReformatBlocks;
 }
 
+bool SyntaxHighlighter::syntaxHighlighterUpToDate() const
+{
+    return d->syntaxInfoUpToDate;
+}
+
+void SyntaxHighlighter::clearExtraFormats(const QTextBlock &block)
+{
+    const int blockLength = block.length();
+    if (block.layout() == nullptr || blockLength == 0)
+        return;
+
+    const QList<QTextLayout::FormatRange> formatsToApply
+        = Utils::filtered(block.layout()->formats(), [](const QTextLayout::FormatRange &r) {
+              return !r.format.property(SemanticHighlight).toBool();
+          });
+
+    bool wasInReformatBlocks = d->inReformatBlocks;
+    d->inReformatBlocks = true;
+    block.layout()->setFormats(formatsToApply);
+
+    document()->markContentsDirty(block.position(), blockLength - 1);
+    d->inReformatBlocks = wasInReformatBlocks;
+}
+
+void SyntaxHighlighter::clearAllExtraFormats()
+{
+    QTextBlock b = document()->firstBlock();
+    while (b.isValid()) {
+        clearExtraFormats(b);
+        b = b.next();
+    }
+}
+
 /* Generate at least n different colors for highlighting, excluding background
  * color. */
 
@@ -735,7 +817,7 @@ QList<QColor> SyntaxHighlighter::generateColors(int n, const QColor &background)
     // Assign a color gradient. Generate a sufficient number of colors
     // by using ceil and looping from 0..step.
     const double oneThird = 1.0 / 3.0;
-    const int step = qRound(ceil(pow(double(n), oneThird)));
+    const int step = qRound(std::ceil(std::pow(double(n), oneThird)));
     result.reserve(step * step * step);
     const int factor = 255 / step;
     const int half = factor / 2;
@@ -762,17 +844,12 @@ QList<QColor> SyntaxHighlighter::generateColors(int n, const QColor &background)
 
 void SyntaxHighlighter::setFontSettings(const FontSettings &fontSettings)
 {
-    Q_D(SyntaxHighlighter);
     d->updateFormats(fontSettings);
 }
-/*!
-    The syntax highlighter is not anymore reacting to the text document if \a noAutmatic is
-    \c true.
-*/
-void SyntaxHighlighter::setNoAutomaticHighlighting(bool noAutomatic)
+
+FontSettings SyntaxHighlighter::fontSettings() const
 {
-    Q_D(SyntaxHighlighter);
-    d->noAutomaticHighlighting = noAutomatic;
+    return d->fontSettings;
 }
 
 /*!
@@ -796,7 +873,7 @@ void SyntaxHighlighter::setDefaultTextFormatCategories()
 void SyntaxHighlighter::setTextFormatCategories(int count,
                                                 std::function<TextStyle(int)> formatMapping)
 {
-    QVector<std::pair<int, TextStyle>> categories;
+    QList<std::pair<int, TextStyle>> categories;
     categories.reserve(count);
     for (int i = 0; i < count; ++i)
         categories.append({i, formatMapping(i)});
@@ -812,23 +889,35 @@ void SyntaxHighlighter::setTextFormatCategories(int count,
 
     \sa setDefaultTextFormatCategories()
 */
-void SyntaxHighlighter::setTextFormatCategories(const QVector<std::pair<int, TextStyle>> &categories)
+void SyntaxHighlighter::setTextFormatCategories(const QList<std::pair<int, TextStyle>> &categories)
 {
-    Q_D(SyntaxHighlighter);
     d->formatCategories = categories;
     const int maxCategory = Utils::maxElementOr(categories, {-1, C_TEXT}).first;
-    d->formats = QVector<QTextCharFormat>(maxCategory + 1);
-    d->updateFormats(TextEditorSettings::fontSettings());
+    d->formats = QList<QTextCharFormat>(maxCategory + 1);
+    d->updateFormats(d->fontSettings);
 }
 
 QTextCharFormat SyntaxHighlighter::formatForCategory(int category) const
 {
-    Q_D(const SyntaxHighlighter);
     QTC_ASSERT(d->formats.size() > category, return QTextCharFormat());
 
     return d->formats.at(category);
 }
 
+QTextCharFormat SyntaxHighlighter::whitespacified(const QTextCharFormat &fmt)
+{
+    QTextCharFormat format = d->whitespaceFormat;
+    format.setBackground(fmt.background());
+    return format;
+}
+
+QTextCharFormat SyntaxHighlighter::asSyntaxHighlight(const QTextCharFormat &fmt)
+{
+    QTextCharFormat format = fmt;
+    format.setProperty(SyntaxHighlight, true);
+    return format;
+}
+
 void SyntaxHighlighter::highlightBlock(const QString &text)
 {
     formatSpaces(text);
@@ -836,9 +925,10 @@ void SyntaxHighlighter::highlightBlock(const QString &text)
 
 void SyntaxHighlighterPrivate::updateFormats(const FontSettings &fontSettings)
 {
+    this->fontSettings = fontSettings;
     // C_TEXT is handled by text editor's foreground and background color,
     // so use empty format for that
-    for (const auto &pair : qAsConst(formatCategories)) {
+    for (const auto &pair : std::as_const(formatCategories)) {
         formats[pair.first] = pair.second == C_TEXT ? QTextCharFormat()
                                                     : fontSettings.toTextCharFormat(pair.second);
     }
@@ -846,5 +936,3 @@ void SyntaxHighlighterPrivate::updateFormats(const FontSettings &fontSettings)
 }
 
 } // namespace TextEditor
-
-#include "moc_syntaxhighlighter.cpp"
